project = "AttiPy"
copyright = "2026, Vegard R. Solum"
author = "Vegard R. Solum"
release = "0.0.1"

extensions = []

templates_path = ["_templates"]
exclude_patterns = []

html_theme = "alabaster"
html_static_path = ["_static"]
